/* Copyright (c) 2015 Yajirushi(Cursor)
 *
 * Released under the MIT license
 * http://opensource.org/licenses/mit-license.php
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#include <Arduino.h>
#include <SD.h>

#ifndef SH_MLCD_MONO
#define SH_MLCD_MONO

//以下、使用する環境に合わせて変更 ==================================

//MLCDのデータシート上の表示領域のサイズ(ピクセル)
//縦横を逆にするとうまく描画できないので注意
//SH_MLCD_MONO_PXWIDTHは必ず8で割り切れる数値であること
#define SH_MLCD_MONO_PXWIDTH 400
#define SH_MLCD_MONO_PXHEIGHT 240

//SPIモドキの持ち時間(ほとんどの場合1usのままで良い)
#define SH_MLCD_MONO_SPIWAIT 1

//以降、変更不要 ====================================================

//MLCDの命令
#define SH_MLCD_MONO_WRITE 0x80
#define SH_MLCD_MONO_COM 0x40
#define SH_MLCD_MONO_CLEAR 0x20
#define SH_MLCD_MONO_DUMMY 0x00

//ピクセルモード
#define SH_MLCD_MONO_BLACK 0
#define SH_MLCD_MONO_ERASE 1
#define SH_MLCD_MONO_INVERT -1

//drawBMPfromSD戻り値
#define ERR_ILLEGAL_SIZE -1
#define ERR_CANNOT_BEGIN -2
#define ERR_NOT_EXISTS -4
#define ERR_CANNOT_OPEN -8
#define ERR_ISNOT_BITMAP -16
#define ERR_IS_OLDBITMAP -17
#define ERR_IS_SPECIALBITMAP -18
#define ERR_COLOR_DEPTH -19
#define ERR_COMPRESSED -20
#define SUCCESS_UPDATE 0
#define SUCCESS_PRINTED 1

/** シャープ製メモリ液晶用直接駆動ライブラリ
 *
 * 秋月電子通商などで販売されている、シャープ製メモリ液晶を直接駆動させるためのライブラリです。@n
 * 「みんなのラボ」から販売されている、SRAMを介すもの(メモリ液晶ブレークアウト基板)では動作しません。@n
 *
 * 2015年時点で販売されている、同じ駆動方式のLCD一覧：(このライブラリで駆動可能)@n
 * LS010B7DH01, LS013B7DH05, LS013B7DH03, LS027B7DH01A, LS044Q7DH01@n
 * (LS012B7DD01, LS013B7DH06は駆動方式が異なるため使用できません)
 *
 * Example:
 * @code
 * //GR-SAKURA Sketch Template V2.07
 * #include <Arduino.h>
 * #include <MsTimer2.h>
 * #include "SH_MEMORYLCD.h"
 * 
 * //COM反転処理間隔(ミリ秒)
 * #define TIMER_VAL_MS 200
 * 
 * //モード待ち時間用(ミリ秒)
 * #define TIMER_VAL_IMGMODE 3500
 * 
 * //MLCD_MONO mlcd(7, 6, 5, 4, 3);
 * MLCD_MONO mlcd(7, 6, 5);
 * 
 * //COM反転処理
 * void invCOM(){
 *     mlcd.invertCOM();
 * }
 * 
 * char imgMode; //ループでのモード判定用
 * unsigned long ti; //時間計測用
 * String modes[] = {"グリッド","図形","文字列","サイン波","画像"};
 * 
 * //セットアップ:最初の1回だけ実行される
 * void setup(){
 *     pinMode(PIN_LED1,OUTPUT); //LED1
 *     pinMode(PIN_LED2,OUTPUT); //LED2
 *     pinMode(PIN_LED3,OUTPUT); //LED3
 *     
 *     Serial.begin(9600);
 *     
 *     pinMode(PIN_SW,INPUT); //押しボタン(青)
 *     
 *     digitalWrite(PIN_LED1, HIGH); //LED1点灯
 *     digitalWrite(PIN_LED2, HIGH); //LED2点灯
 *     digitalWrite(PIN_LED3, HIGH); //LED3点灯
 *     
 *     while(digitalRead(PIN_SW) == 1); //押しボタン(青)が押されるまで待ち続ける
 *     
 *     digitalWrite(PIN_LED2, LOW); //LED2消灯
 *     
 *     mlcd.init(); //液晶初期化
 *     
 *     //MsTimer2にinvCOMを登録してタイマー開始
 *     MsTimer2::set(TIMER_VAL_MS, invCOM);
 *     MsTimer2::start();
 *     
 *     imgMode = 0;
 *     
 *     digitalWrite(PIN_LED3, LOW); //LED3消灯
 *     
 *     delay(1000);
 * }
 * 
 * void loop(){
 *     
 *     //0から4の中で循環させる
 *     imgMode = (imgMode + 1) % 5;
 *     
 *     //モードに応じて処理を変える
 *     switch(imgMode){
 *         //モード0:グリッド
 *         case 0:{
 *             //時間計測
 *             ti = millis();
 *             
 *             //垂直線を等間隔に(描画はあとで)
 *             for(int x=10; x<SH_MLCD_MONO_PXWIDTH; x+=20){
 *                 mlcd.drawLine(x, 0, x, SH_MLCD_MONO_PXHEIGHT-1, SH_MLCD_MONO_BLACK, false);
 *             }
 *             
 *             //水平線を等間隔に(描画はあとで)
 *             for(int y=10; y<SH_MLCD_MONO_PXHEIGHT; y+=20){
 *                 mlcd.drawLine(0, y, SH_MLCD_MONO_PXWIDTH-1, y, SH_MLCD_MONO_BLACK, false);
 *             }
 *             
 *             //一気に描画(内部配列を液晶に転送)
 *             mlcd.writeArrayA();
 *             
 *             //時間計測
 *             ti = millis() - ti;
 *             
 *             break;
 *         }
 *         //モード1:図形
 *         case 1:{
 *             //時間計測
 *             ti = millis();
 *             
 *             //四角形(即時描画)
 *             mlcd.drawRect(15, 15, 20, 20, SH_MLCD_MONO_BLACK, true);
 *             mlcd.drawRect2(45, 15, 65, 35, SH_MLCD_MONO_BLACK, true);
 *             mlcd.fillRect(75, 15, 20, 20, SH_MLCD_MONO_BLACK, true);
 *             mlcd.fill2Point(105, 15, 125, 35, SH_MLCD_MONO_BLACK, true);
 *             
 *             //円(即時描画)
 *             mlcd.drawCircle(25, 55, 10, SH_MLCD_MONO_BLACK, true);
 *             mlcd.fillCircle(55, 55, 10, SH_MLCD_MONO_BLACK, true);
 *             
 *             //四角形(反転)
 *             mlcd.fill2Point(19, 19, 95, 95, SH_MLCD_MONO_INVERT, true);
 *             
 *             //時間計測
 *             ti = millis() - ti;
 *             
 *             break;
 *         }
 *         //モード2:文字列
 *         case 2:{
 *             //時間計測
 *             ti = millis();
 *             
 *             //文字列表示(即時描画)
 *             String str1 = "writeString(:String) zoom x1";
 *             mlcd.ws(str1, 20, 20);
 *             
 *             //文字列表示(即時描画)
 *             char str2[] = "writeString(:char*) zoom x1";
 *             mlcd.writeString(str2, 20, 40);
 *             
 *             //日本語表示(改行あり)
 *             str1 = "日本語メッセージも余裕。\nただしメモリをたくさん食います。";
 *             mlcd.writeString(str1, 20, 60);
 *             
 *             //大きい文字
 *             char str3[] = "Large Text! 大きい文字も余裕。";
 *             mlcd.ws(str3, 10, 160, 2, SH_MLCD_MONO_BLACK, true);
 *             
 *             //時間計測
 *             ti = millis() - ti;
 *             
 *             break;
 *         }
 *         //モード3:サイン波
 *         case 3:{
 *             //時間計測
 *             ti = millis();
 *             
 *             //3つの120度ずれたサイン波を描画(即時描画)
 *             float deg120 = 2.09439510239, sH = 65;
 *             float y1=0, y2=0, y3=0;
 *             int gW = 200;
 *             for(float x=100; x<gW; x+=0.15){
 *                 float x0 = x/18.5;
 *                 y1 = sH * sin(x0 - deg120) + sH;
 *                 y2 = sH * sin(x0) + sH;
 *                 y3 = sH * sin(x0 + deg120) + sH;
 *                 mlcd.pixel(x, y1+55);
 *                 mlcd.pixel(x, y2+55);
 *                 mlcd.pixel(x, y3+55);
 *             }
 *             
 *             //時間計測
 *             ti = millis() - ti;
 *             
 *             break;
 *         }
 *         //モード4:画像
 *         case 4:{
 *             //時間計測
 *             ti = millis();
 *             
 *             //SDカード内のビットマップ画像データ(カラーをディザリング)を表示する
 *             String fileName = "c.bmp";
 *             mlcd.drawBMPfromSD(fileName, true, 0, 0);
 *             
 *             //時間計測
 *             ti = millis() - ti;
 *             
 *             break;
 *         }
 *     }
 *     
 *     delay(TIMER_VAL_IMGMODE);
 *     
 *     //文字表示部を消す
 *     mlcd.fill2Point(80, 80, 319, 180, SH_MLCD_MONO_ERASE, true);
 *     mlcd.drawRect2(80, 80, 319, 180, SH_MLCD_MONO_BLACK, true);
 *     
 *     //モード名と処理時間を表示
 *     mlcd.ws(modes[imgMode], 90, 85);
 *     char msg[100];
 *     sprintf(msg, "処理時間 = %ld millis", ti);
 *     mlcd.ws(msg, 90, 115);
 *     
 *     delay(TIMER_VAL_IMGMODE);
 *     
 *     //表示を消して、RAMもクリア
 *     mlcd.cls();
 *     mlcd.cla();
 *     
 *     if(digitalRead(PIN_SW) == 0){
 *         digitalWrite(PIN_LED1, LOW); //LED1消灯
 *         exit(1);
 *     }
 * }
 * @endcode
 */
class MLCD_MONO{
public:
	/** コンストラクタ:MLCD_MONOクラスのインスタンスを作成します
	 *
	 *  @param vSCK メモリ液晶のSCK端子に出力するピン番号を指定
	 *  @param vSI メモリ液晶のSI端子に出力するピン番号を指定
	 *  @param vCS メモリ液晶のSCS端子に出力するピン番号を指定
	 *  @param vCOM メモリ液晶のCOM端子に出力するピン番号を指定
	 *  @param vDISP メモリ液晶のDISP端子に出力するピン番号を指定
	 */
	MLCD_MONO(unsigned char vSCK, unsigned char vSI, unsigned char vCS, unsigned char vCOM, unsigned char vDISP);
	
	/** コンストラクタ:MLCD_MONOクラスのインスタンスを作成します(オーバーロード)
	 *
	 *  @param vSCK メモリ液晶のSCK端子に出力するピン番号を指定
	 *  @param vSI メモリ液晶のSI端子に出力するピン番号を指定
	 *  @param vCS メモリ液晶のSCS端子に出力するピン番号を指定
	 */
	MLCD_MONO(unsigned char vSCK, unsigned char vSI, unsigned char vCS);
	
	/** デストラクタ(ほとんどの場合、ユーザーが明示的に呼ぶ必要はありません)
	 */
	~MLCD_MONO();

private:
	//ディスプレイの幅、高さ
	unsigned short _width, _height;
	
	//一行のデータ量(_width / 8)
	unsigned char _bytes;
	
	//画素状態記憶用配列
	//最大12KByte(400 / 8 * 240 = 12,000)
	char *R;
	
	//通信に使用するピン(SPIモドキ)
	unsigned char pinSCK, pinSI, pinSCS;
	
	//外部信号用ピン
	unsigned char pinCOM, pinDISP;
	bool useEcom, useEdisp;
	
	//COM信号状態
	bool comFlag;
	
	//命令送信フラグ
	bool busySPI;
	
	//SPIモドキのコマンド送出
	void send_mode(unsigned char val);
	void send_addr(unsigned char val);
	void send_data(unsigned char val);
	
	//ファイルの部分読み取り
	short fr2(File srcf);
	long fr4(File srcf);
	
public:
	//ユーティリティ ==========================================================
	/** 初期化:使用する配列や端子の状態を初期化します
	 * @warning インスタンス1つにつき1回しか呼ばないこと
	 *
	 * @param initPIN コンストラクタで指定したピンの初期化を行うかどうか(デフォルト=true)
	 */
	void init(bool initPIN=true);
	
	/** 液晶の表示をクリアする:内部画素記憶配列はクリアされない
	 */
	void clear();

	/** 液晶の表示をクリアする:内部画素記憶配列はクリアされない(clearと同一)
	 */
	void cls();
	
	/** 内部画素記憶配列をクリアする:表示はクリアされない
	 */
	void clearArray();
	
	/** 内部画素記憶配列をクリアする:表示はクリアされない(clearArrayと同一)
	 */
	void cla();
	
	/** DISP端子をONにする:外部DISP信号を使用しない場合は何も起きない
	 */
	void dispON();

	/** DISP端子をOFFにする:外部DISP信号を使用しない場合は何も起きない
	 */
	void dispOFF();
	
	/** 液晶寿命を延ばすためにCOM信号を反転させる@n
	 * @warning クラスの外でタイマー処理(MsTimer2など)を使って定期的にこの関数を呼ぶこと
	 */
	void invertCOM();
	
	//画素配列操作 ============================================================
	/** 内部画素記憶配列の特定の行にデータを上書きする
	 *
	 *  @param ln 上書きする行(0 から 高さ-1 まで)
	 *  @param data 上書きするデータ
	 */
	void updateArray1(unsigned char ln, const char *data);
	
	/** 内部画素記憶配列の特定の範囲にデータを上書きする
	 *
	 *  @param sn 上書きを開始する行(0 から 高さ-1 まで)
	 *  @param en 上書きが終了する行(0 から 高さ-1 まで)
	 *  @param data 上書きするデータ
	 */
	void updateArrayN(unsigned char sn, unsigned char en, const char *data);
	
	/** 内部画素記憶配列を全て上書きする
	 *
	 *  @param data 上書きするデータ
	 */
	void updateArrayA(const char *data);
	
	/** 内部画素記憶配列の特定の行をメモリ液晶に表示する
	 *
	 *  @param ln 表示する行(0 から 高さ-1 まで)
	 */
	void writeArray1(unsigned char ln);
	
	/** 内部画素記憶配列の特定の範囲をメモリ液晶に表示する
	 *
	 *  @param sn 表示したい範囲の開始行(0 から 高さ-1 まで)
	 *  @param sn 表示したい範囲の終了行(0 から 高さ-1 まで)
	 */
	void writeArrayN(unsigned char sn, unsigned char en);
	
	/** 内部画素記憶配列を全てメモリ液晶に表示する
	 */
	void writeArrayA();
	
	/** データをメモリ液晶の特定行に表示する(内部画素記憶配列を介さない)
	 *
	 *  @param ln 表示する行(0 から 高さ-1 まで)
	 *  @param data 表示するデータ
	 */
	void write1(unsigned char ln, const char *data);
	
	/** データをメモリ液晶の特定の範囲に表示する(内部画素記憶配列を介さない)
	 *
	 *  @param sn 表示したい範囲の開始行(0 から 高さ-1 まで)
	 *  @param en 表示したい範囲の終了行(0 から 高さ-1 まで)
	 *  @param data 表示するデータ
	 */
	void writeN(unsigned char sn, unsigned char en, const char *data);
	
	/** データをメモリ液晶全域に表示する(内部画素記憶配列を介さない)
	 *
	 *  @param data 表示するデータ
	 */
	void writeA(const char *data);
	
	//図形描画 ================================================================
	/** メモリ液晶の特定の座標に1ピクセルのドットを表示する
	 *
	 *  @param x X座標(0 から 幅-1 まで)
	 *  @param y Y座標(0 から 高さ-1 まで)
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 */
	void pixel(unsigned short x, unsigned short y, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** メモリ液晶に線を表示する
	 *
	 *  @param fromX 始点のX座標(0 から 幅-1 まで)
	 *  @param fromY 始点のY座標(0 から 高さ-1 まで)
	 *  @param toX 終点のX座標(0 から 幅-1 まで)
	 *  @param toY 終点のY座標(0 から 高さ-1 まで)
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 */
	void drawLine(unsigned short fromX, unsigned short fromY, unsigned short toX, unsigned short toY, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** メモリ液晶に四角形を表示する
	 *
	 *  @param left 画面左端からの距離(0 から 幅-1 まで)
	 *  @param top 画面上端からの距離(0 から 高さ-1 まで)
	 *  @param width 四角形の幅
	 *  @param height 四角形の高さ
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 */
	void drawRect(unsigned short left, unsigned short top, unsigned short width, unsigned short height, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** メモリ液晶に四角形を表示する(対角点指定)
	 *
	 *  @param x1 四角形の左上の点のX座標(0 から 幅-1 まで)
	 *  @param y1 四角形の左上の点のY座標(0 から 高さ-1 まで)
	 *  @param x2 四角形の右下の点のX座標(0 から 幅-1 まで)
	 *  @param y2 四角形の右下の点のY座標(0 から 高さ-1 まで)
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 */
	void drawRect2(unsigned short x1, unsigned short y1, unsigned short x2, unsigned short y2, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** メモリ液晶に塗りつぶされた四角形を表示する
	 *
	 *  @param left 画面左端からの距離(0 から 幅-1 まで)
	 *  @param top 画面上端からの距離(0 から 高さ-1 まで)
	 *  @param width 四角形の幅
	 *  @param height 四角形の高さ
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 */
	void fillRect(unsigned short left, unsigned short top, unsigned short width, unsigned short height, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** メモリ液晶に塗りつぶされた四角形を表示する(対角点指定)
	 *
	 *  @param x1 四角形の左上の点のX座標(0 から 幅-1 まで)
	 *  @param y1 四角形の左上の点のY座標(0 から 高さ-1 まで)
	 *  @param x2 四角形の右下の点のX座標(0 から 幅-1 まで)
	 *  @param y2 四角形の右下の点のY座標(0 から 高さ-1 まで)
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 */
	void fill2Point(unsigned short x1, unsigned short y1, unsigned short x2, unsigned short y2, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** メモリ液晶に円を表示する
	 *
	 *  @param centerX 円の中心のX座標(0 から 幅-1 まで)
	 *  @param centerY 円の中心のY座標(0 から 高さ-1 まで)
	 *  @param r 円の半径
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 */
	void drawCircle(unsigned short centerX, unsigned short centerY, unsigned short r, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** メモリ液晶に塗りつぶされた円を表示する
	 *
	 *  @param centerX 円の中心のX座標(0 から 幅-1 まで)
	 *  @param centerY 円の中心のY座標(0 から 高さ-1 まで)
	 *  @param r 円の半径
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): fillCircleは描画方法が特殊なためうまく描画できません@n
	 *
	 *  @param immidiate 即時描画させる
	 */
	void fillCircle(unsigned short centerX, unsigned short centerY, unsigned short r, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	//画像の描画 ==============================================================
	/** メモリ液晶にモノクロ2値のデータ列を表示する
	 *
	 *  @param data 表示するデータ列
	 *  @param left 画面左端からの距離(0 から 幅-1 まで)
	 *  @param top 画面上端からの距離(0 から 高さ-1 まで)
	 *  @param imgWidth 四角形の幅
	 *  @param imgHeight 四角形の高さ
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): fillCircleは描画方法が特殊なためうまく描画できません@n
	 *
	 *  @param immidiate 即時描画させる
	 *
	 *  @retval ERR_ILLEGAL_SIZE(-1) 
	 *  @retval SUCCESS_UPDATE(0) 内部配列を更新した
	 *  @retval SUCCESS_PRINTED(1) 内部配列を更新した上、メモリ液晶に表示した
	 */
	char drawMonoData(const char *data, unsigned short left, unsigned short top, unsigned short imgWidth, unsigned short imgHeight, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** SDカードから読み取ったBMP画像をメモリ液晶に表示する
	 *
	 *  @param path BMP画像のパス文字列(char*)
	 *  @param left 画面左端からの距離(0 から 幅-1 まで)
	 *  @param top 画面上端からの距離(0 から 高さ-1 まで)
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 *
	 *  @retval ERR_CANNOT_BEGIN(-2) SD.begin()がうまくいかなかった
	 *  @retval ERR_NOT_EXISTS(-4) 指定されたパスが存在しない
	 *  @retval ERR_CANNOT_OPEN(-8) ファイルは存在するが、開くことができない
	 *  @retval ERR_ISNOT_BITMAP(-16) 対象ファイルがビットマップファイルではない
	 *  @retval ERR_IS_OLDBITMAP(-17) 対象ファイルが古い形式のビットマップファイルなので解析できない
	 *  @retval ERR_IS_SPECIALBITMAP(-18) 対象ファイルが特殊形式のビットマップファイルなので解析できない
	 *  @retval ERR_COLOR_DEPTH(-19) 対象ファイルの色深度が1bit,24bit,32bit以外であるため解析できない
	 *  @retval ERR_COMPRESSED(-20) 対象ファイルが圧縮されているので解析できない
	 *  @retval SUCCESS_UPDATE(0) 対象ファイルの読み込みに成功し、内部配列を更新した
	 *  @retval SUCCESS_PRINTED(1) 対象ファイルの読み込みに成功し、内部配列を更新した上、メモリ液晶に表示した
	 */
	char drawBMPfromSD(char *path, bool BtoW, unsigned short left, unsigned short top, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** SDカードから読み取ったBMP画像をメモリ液晶に表示する (オーバーロード)
	 *
	 *  @param path BMP画像のパス文字列(String)
	 *  @param left 画面左端からの距離(0 から 幅-1 まで)
	 *  @param top 画面上端からの距離(0 から 高さ-1 まで)
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 *
	 *  @retval ERR_CANNOT_BEGIN(-2) SD.begin()がうまくいかなかった
	 *  @retval ERR_NOT_EXISTS(-4) 指定されたパスが存在しない
	 *  @retval ERR_CANNOT_OPEN(-8) ファイルは存在するが、開くことができない
	 *  @retval ERR_ISNOT_BITMAP(-16) 対象ファイルがビットマップファイルではない
	 *  @retval ERR_IS_OLDBITMAP(-17) 対象ファイルが古い形式のビットマップファイルなので解析できない
	 *  @retval ERR_IS_SPECIALBITMAP(-18) 対象ファイルが特殊形式のビットマップファイルなので解析できない
	 *  @retval ERR_COLOR_DEPTH(-19) 対象ファイルの色深度が1bit,24bit,32bit以外であるため解析できない
	 *  @retval ERR_COMPRESSED(-20) 対象ファイルが圧縮されているので解析できない
	 *  @retval SUCCESS_UPDATE(0) 対象ファイルの読み込みに成功し、内部配列を更新した
	 *  @retval SUCCESS_PRINTED(1) 対象ファイルの読み込みに成功し、内部配列を更新した上、メモリ液晶に表示した
	 */
	char drawBMPfromSD(String path, bool BtoW, unsigned short left, unsigned short top, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	//文字の描画 ==============================================================
	/** メモリ液晶に文字を表示する
	 *
	 *  @param s 表示する文字
	 *  @param x テキストを囲う四角形の左上のX座標
	 *  @param y テキストを囲う四角形の左上のY座標
	 *  @param zoom 拡大倍率
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 *
	 *  @retval -1 表示できなかった
	 *  @retval 0以上 文字のバイトサイズ
	 */
	char drawChar(const char *s, unsigned short x, unsigned short y, unsigned char zoom, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** メモリ液晶に拡大倍率を指定して文字列を表示する
	 *
	 *  @param str 表示する文字列(char*)
	 *  @param x テキストを囲う四角形の左上のX座標
	 *  @param y テキストを囲う四角形の左上のY座標
	 *  @param zoom 拡大倍率
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 *
	 *  @retval -1 表示できなかった
	 *  @retval 1 表示できた
	 */
	char writeString(const char *str, unsigned short x, unsigned short y, unsigned char zoom, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** メモリ液晶に拡大倍率を指定して文字列を表示する(オーバーロード)
	 *
	 *  @param str 表示する文字列(String)
	 *  @param x テキストを囲う四角形の左上のX座標
	 *  @param y テキストを囲う四角形の左上のY座標
	 *  @param zoom 拡大倍率
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 *
	 *  @retval -1 表示できなかった
	 *  @retval 1 表示できた
	 */
	char writeString(const String str, unsigned short x, unsigned short y, unsigned char zoom, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** メモリ液晶に文字列を表示する(オーバーロード)
	 *
	 *  @param str 表示する文字列(char*)
	 *  @param x テキストを囲う四角形の左上のX座標
	 *  @param y テキストを囲う四角形の左上のY座標
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 *
	 *  @retval -1 表示できなかった
	 *  @retval 1 表示できた
	 */
	char writeString(const char *str, unsigned short x, unsigned short y, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** メモリ液晶に文字列を表示する(オーバーロード)
	 *
	 *  @param str 表示する文字列(String)
	 *  @param x テキストを囲う四角形の左上のX座標
	 *  @param y テキストを囲う四角形の左上のY座標
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 *
	 *  @retval -1 表示できなかった
	 *  @retval 1 表示できた
	 */
	char writeString(const String str, unsigned short x, unsigned short y, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** メモリ液晶に拡大倍率を指定して文字列を表示する(writeStringのエイリアス)
	 *
	 *  @param str 表示する文字列(char*)
	 *  @param x テキストを囲う四角形の左上のX座標
	 *  @param y テキストを囲う四角形の左上のY座標
	 *  @param zoom 拡大倍率
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 *
	 *  @retval -1 表示できなかった
	 *  @retval 1 表示できた
	 */
	char ws(const char *str, unsigned short x, unsigned short y, unsigned char zoom, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** メモリ液晶に拡大倍率を指定して文字列を表示する(オーバーロード)(writeStringのエイリアス)
	 *
	 *  @param str 表示する文字列(String)
	 *  @param x テキストを囲う四角形の左上のX座標
	 *  @param y テキストを囲う四角形の左上のY座標
	 *  @param zoom 拡大倍率
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 *
	 *  @retval -1 表示できなかった
	 *  @retval 1 表示できた
	 */
	char ws(const String str, unsigned short x, unsigned short y, unsigned char zoom, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** メモリ液晶に文字列を表示する(オーバーロード)(writeStringのエイリアス)
	 *
	 *  @param str 表示する文字列(char*)
	 *  @param x テキストを囲う四角形の左上のX座標
	 *  @param y テキストを囲う四角形の左上のY座標
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 *
	 *  @retval -1 表示できなかった
	 *  @retval 1 表示できた
	 */
	char ws(const char *str, unsigned short x, unsigned short y, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
	
	/** メモリ液晶に拡大倍率を指定して文字列を表示する(writeStringのエイリアス)
	 *
	 *  @param str 表示する文字列(String)
	 *  @param x テキストを囲う四角形の左上のX座標
	 *  @param y テキストを囲う四角形の左上のY座標
	 *  @param zoom 拡大倍率
	 *  @param mode 表示モード
	 *
	 *  @note modeの一覧:
	 *   SH_MLCD_MONO_BLACK (0): 黒く表示する@n
	 *   SH_MLCD_MONO_ERASE (1): 白く表示する@n
	 *   SH_MLCD_MONO_INVERT (-1): 画素を逆転させる(黒なら白、白なら黒)@n
	 *
	 *  @param immidiate 即時描画させる
	 *
	 *  @retval -1 表示できなかった
	 *  @retval 1 表示できた
	 */
	char ws(const String str, unsigned short x, unsigned short y, char mode=SH_MLCD_MONO_BLACK, bool immidiate=true);
};

#endif